library(Rlabkey)


#' Check if a Path Exists on the Labkey Server
#'
#' This function verifies whether a specified file or directory exists on the Labkey server
#' using the WebDAV API. It returns a list containing the existence status and related information.
#'
#' @param folder_path A character string specifying the path to the project folder on the Labkey server.
#' @param remote_file_path (optional) A character string specifying the relative path of the file within the project folder.
#'        Can be an empty string or NULL if checking the folder itself.
#'
#' @return A list containing the following elements:
#' \item{exists}{Logical, whether the specified path exists on the Labkey server.}
#' \item{url}{Character string, the base URL of the Labkey server (NULL if the path does not exist).}
#' \item{project}{Character string, the folder path of the project (NULL if the path does not exist).}
#' \item{filename}{Character string, the relative file path within the project (NULL if the path does not exist).}
#'
#' @details
#' This function leverages the `labkey.webdav.pathExists` method to check whether the specified path exists
#' on the server. If the path exists, it returns metadata about the location; otherwise, it returns NULL values.
#'
#' @examples
#' check_path_exists("/myProject", "data/file.csv")  # Check if a specific file exists
#' check_path_exists("/myProject", ".")  # Check if the project folder exists
check_path_exists <- function(folder_path, remote_file_path) {
  if (labkey.webdav.pathExists(baseUrl = labkey.getBaseUrl(), folderPath = folder_path, remoteFilePath = remote_file_path)) {
    return(list(exists = TRUE, url = labkey.getBaseUrl(), project = folder_path, filename = remote_file_path))
  }
  return(list(exists = FALSE, url = NULL, project = NULL, filename = NULL))
}




#' Download an Entire Project from the Labkey Server
#'
#' This function downloads an entire project folder from the Labkey server to the local system.
#' The folder is saved under the user's home directory, and if the local directory does not exist, it will be created.
#'
#' @param folder_path A character string specifying the path to the project folder on the Labkey server.
#'
#' @return No return value. Downloads the entire project folder to the local system and prints a message indicating the download location.
#'
#' @details
#' The function uses the `labkey.webdav.downloadFolder` method to download the project folder,
#' including all subdirectories and files. The folder is saved under the user's home directory, in a subdirectory named `_webdav`.
#' If the directory structure does not exist, it will be created.
#'
#' @examples
#' download_project("/myProject")  # Downloads the entire project folder from the Labkey server
download_project <- function(folder_path) {
  local_base_dir <- file.path(Sys.getenv("HOME"))
  if (!dir.exists(local_base_dir)) {
    dir.create(local_base_dir, recursive = TRUE)
  }
  message("Downloading folder to ", local_base_dir)
  resp <- labkey.webdav.downloadFolder(
    baseUrl = labkey.getBaseUrl(),
    localBaseDir = local_base_dir,
    folderPath = folder_path,
    remoteFilePath = ".",
    overwriteFiles = TRUE,
    mergeFolders = TRUE,
    fileSet = "@files"
  )
  message("response : ", resp)
}




#' Download a Specific File from the Labkey Server
#'
#' This function downloads a specific file from the Labkey server to the local system.
#' The file is saved under the user's home directory, with the folder structure preserved.
#'
#' @param folder_path A character string specifying the path to the project folder on the Labkey server.
#' @param remote_file_path A character string specifying the relative path of the file within the project folder.
#'
#' @return No return value. Downloads the specified file to the local system and prints a message indicating the download location.
#'
#' @details
#' The function uses the `labkey.webdav.get` method to download the specified file from the Labkey server.
#' The file is saved under the user's home directory in a subdirectory named `_webdav`, maintaining the folder structure.
#' If the `_webdav` directory does not exist, it will be created.
#'
#' @examples
#' download_file("/myProject", "data/file.csv")  # Downloads a specific file from the project folder on the Labkey server
download_file <- function(folder_path, remote_file_path) {
  local_base_dir <- file.path(Sys.getenv("HOME"), "_webdav")
  local_file_path <- file.path(local_base_dir, folder_path, remote_file_path)
  if (!dir.exists(local_base_dir)) {
    dir.create(local_base_dir, recursive = TRUE)
    message("Created local directory: ", local_base_dir)
  }
  message("Downloading file to ", local_file_path)
  resp <- labkey.webdav.get(
    baseUrl = labkey.getBaseUrl(),
    folderPath = folder_path,
    remoteFilePath = remote_file_path,
    localFilePath = local_file_path,
    overwrite = TRUE,
    fileSet = "@files"
  )
  message("response : ", resp)
}




#' Fetch files from Labkey Server
#'
#' This function retrieves data from Labkey server, either as a directory or a specific file.
#' It first checks whether the specified path exists on the server, and if not, it stops execution with an error message.
#' If the path exists, it downloads the entire project folder or the specified file.
#'
#' @param project_name A character string specifying the name of the project to fetch (need to have access to the project).
#' @param remote_file_path (optional) A character string specifying the relative path of the file within the project.
#'        If NULL, the entire project directory will be downloaded. Default is NULL.
#'
#' @return Prints a message indicating the type of data being fetched and the server details.
#'
#' @details
#' The function performs the following steps:
#' - Constructs the folder path using the project name `check_path_exists`.
#' - Checks if the specified file or directory exists on the configured server `check_path_exists`.
#' - If the path exists, prints a message and downloads the file or directory `download_project`, `download_file`.
#' - Stops with an error message if the path does not exist.
#'
#' @examples
#' fetch_data("myProject")  # Fetches the entire project directory
#' fetch_data("myProject", "data/file.csv")  # Fetches a specific file within the project
fetch_data <- function(project_name, remote_file_path = NULL) {
  folder_path <- paste0("/", project_name)
  path_info <- check_path_exists(folder_path, ifelse(is.null(remote_file_path), ".", remote_file_path))

  if (!path_info$exists) {
    stop("The specified path does not exist on Labkey server.")
  }

  message("Fetching ", ifelse(is.null(remote_file_path), "directory", "file"), " from: ",
          path_info$url, " ", path_info$project,
          if (!is.null(remote_file_path)) paste0(" ", path_info$filename))

  # fetch data from Labkey Server
  tryCatch({
    if (is.null(remote_file_path)) {
      download_project(project_name)
    } else {
      download_file(folder_path, remote_file_path)
    }
  }, error = function(e) {
    message("Error during data download: ", e$message)
  })
}



#' Fetch Files or Directories from Labkey Server
#'
#' This function serves as a wrapper to fetch files or directories from a Labkey server
#' by calling the `fetch_data` function.
#' @param projectName A character string specifying the name of the project to fetch from the Labkey server.
#' @param remoteFilePath (optional) A character string specifying the relative path of the file within the project.
#' @return Prints a success message after fetching the files or directories.
#' @details
#' This function is a simple wrapper of `fetch_data()` to streamline file and directory retrieval
#' from the Labkey server.
#' @export
fetch <- function(projectName, remoteFilePath=NULL) {
  project_name <- projectName
  remote_file_path <- remoteFilePath

  fetch_data(project_name, remote_file_path)
  message("Data fetched successfully.")
}
