library(readxl)
library(Rlabkey)

#' Helper function to upload assay data
#'
#' This function is intended to upload assay data.
#'
#' @param df A data frame containing the assay data to upload.
#' @param folder_path The path to the folder where the data should be uploaded.
#' @param target_name The name of the assay to upload data for.
#' @return A message indicating that the upload functionality is not supported yet.
#' @details
#' Uses Rlabkey API method `labkey.saveBatch` to create assay and load dataframe into it.
#'
#' @note This function does nothing currently and outputs a message about its deactivation.
upload_assay_data <- function(df, folder_path, target_name) { # TODO ADD ASSAY UPLOAD SUPPORT
  message("Assay data upload is not supported yet.")

  # bprops <- list(LabNotes = "this is a simple demo")
  # bpl <- list(name = paste("Batch ", as.character(date())), properties = bprops)
  # rpl <- list(name = paste("Assay Run ", as.character(date())))
  # message("Uploading assay data...")
  # assay_info <- labkey.saveBatch(
  #   baseUrl = labkey.getBaseUrl(),
  #   folderPath = folder_path,
  #   assayName = target_name,
  #   resultDataFrame = df,
  #   batchPropertyList = bpl,
  #   runPropertyList = rpl
  # )
  #
  # message("Assay info : ", assay_info)
  # message("Assay data uploaded.")
}

#' Create and Upload a Dataset to LabKey Server
#'
#' Creates a new domain on the LabKey server and uploads the given dataframe to it.
#' The function uses the "StudyDatasetVisit" domain kind and is suitable for uploading
#' structured data related to studies.
#'
#' @param df A data frame containing the data to be uploaded.
#' @param folder_path A string specifying the project folder path on the LabKey server.
#' @param schema_name A string specifying the schema name (usually "study").
#' @param query_name A string specifying the name of the dataset on the server.
#' @details
#' Uses Rlabkey API method `labkey.domain.createAndLoad` to create dataset and load dataframe into it.
#'
#'
#' @note
#' WARNING : The dataframe passed as argument MUST contain the following columns : 'ParticipantId', 'SequenceNum'
#'
#' @return None. Prints messages indicating the creation and upload status.
create_and_upload_dataset <- function(df, folder_path, schema_name, query_name) {
  message("Creating domain: ", query_name)

  # Create the domain and upload the dataframe
  resp <- labkey.domain.createAndLoad(
    baseUrl = labkey.getBaseUrl(),
    domainKind = "StudyDatasetVisit",
    folderPath = folder_path,
    schemaName = schema_name,
    name = query_name,
    df = df,
    description = ""
  )

  message("Domain created and data frame uploaded successfully.")
  message(paste("Data uploaded to", folder_path, ", schema:", schema_name, ", query:", query_name))
}


#' Upload Data to LabKey Server
#'
#' Uploads a data file (CSV, TSV, XLSX, or XLS) to the LabKey server as either an assay or a dataset.
#' The function reads the file based on its extension and then uploads it using appropriate functions
#' depending on the type (assay or dataset).
#'
#' @param project_name A string specifying the project name on the LabKey server.
#' @param file_path A string specifying the local file path to the data file.
#' @param target_name A string specifying the target name on the server.
#' @param is_assay A logical value indicating whether the data is an assay.
#' @param is_dataset A logical value indicating whether the data is a dataset.
#'
#' @details
#' This method is used to read the file with valid extension, and based on parameters upload data :
#' Assay : (Not supported in v1.0.0) using `upload_assay_data` method.
#' Dataset : usin `create_and_upload_dataset` method.
#'
#' @return None. Prints a message indicating success or failure.
upload_data <- function(project_name, file_path, target_name, is_assay, is_dataset) {
  # Validate file extension
  file_ext <- tools::file_ext(file_path)

  # Load the data based on file extension
  df <- switch(file_ext,
               csv = read.csv(file_path, header = TRUE, sep = ","),
               tsv = read.csv(file_path, header = TRUE, sep = "\t"),
               xlsx = read_excel(file_path),
               xls = read_excel(file_path),
               stop("Unsupported file type: ", file_ext))

  # Define folder path
  folder_path <- project_name

  # Upload data based on the specified type
  tryCatch({
    if (is_assay) {
      upload_assay_data(df, folder_path, target_name)
      message("Assay data uploaded successfully to project: ", project_name)
    } else if (is_dataset) {
      schema_name <- "study"
      query_name <- target_name
      create_and_upload_dataset(df, folder_path, schema_name, query_name)
      message("Dataset uploaded successfully to project: ", project_name)
    } else {
      stop("Invalid data type: must specify either assay or dataset.")
    }
  }, error = function(e) {
    message("An error occurred during data upload: ", conditionMessage(e))
  })
}





#' Push Data to LabKey Server
#'
#' Uploads data or directories to the LabKey server with specified project and target names.
#' The function ensures that input parameters are valid and either an assay or a dataset
#' is selected (but not both). In the case of invalid inputs, a detailed message is displayed.
#'
#' @param projectName A string specifying the project name.
#' @param filePath A string specifying the local path to the file to be uploaded.
#' @param targetName A string specifying the target name on the server.
#' @param dirPath (optional) string specifying the directory path (default is NULL).
#' @param isAssay A logical value indicating whether the data is an assay (default is FALSE).
#' @param isDataset A logical value indicating whether the data is a dataset (default is FALSE).
#' @return None. Prints a message indicating success or an error.
#'
#' @details
#' This function verifies and validates inputs, and acts as a wrapper for the `upload_data` method.
#'
#' @export
push <- function(projectName, filePath, targetName, dirPath = NULL, isAssay = FALSE, isDataset = FALSE) {

  # Validate inputs
  if (is.null(filePath) || is.null(projectName) || is.null(targetName) || !xor(isAssay, isDataset)) {
    message("Problem found in inputs.")
    message(paste("Inputs: ",
                  "\nproject_name =", projectName, ", ",
                  "\nfile_path =", filePath, ", ",
                  "\ntarget_name =", targetName, ", ",
                  "\nis_assay =", isAssay, ", ",
                  "\nis_dataset =", isDataset))
    return()
  }

  # Push data to LabKey server
  tryCatch({
    upload_data(projectName, filePath, targetName, isAssay, isDataset)
    message("Data uploaded successfully.")
  }, error = function(e) {
    message("Error during data upload: ", e$message)
  })
}
